#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <libxml/xpath.h>
#include <popt.h>
#include <libharkio3.h>
#include <math.h>
#include "harktool.h"

float calcZValue(int z_unit, float re, float im);
int writeCoordinateFromTF(const char *outfile, int z_axis, harkio_TransferFunction* tf, enum TFType tf_type, bool ifft);
int writeCoordinateInternal(const char* outfile, int z_axis, harkio_Matrix** matdata, const harkio_Positions* poses,
        int num_mic, int num_x, float x_weight, const float* x_offset, float y_weight, const float* z_offset);
void calcIFFTMatrixData(harkio_Matrix** freqdata, harkio_Matrix** timedata, int num_src, int num_mic, int num_fft);

#define DISP_REAL   1
#define DISP_AMP    2
#define DISP_DB     3
#define DISP_PHASE  4

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-zip2coord\n");
    printf("\n");
    printf("Description:\n");
    printf("  Coordinate from transfer function\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int main(int argc, char **argv) {
    char* z_axisStr = "amp";
    int z_axis = DISP_AMP;
    char* tf_typeStr = "LOC";
    enum TFType tf_type = LOC;
    char* logLevel = "E";
    int ifft = false;

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "type", 't',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &tf_typeStr, 't',
      "Type of Transfer function", "LOC|SEP" },
    { "z_axis", 'z',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &z_axisStr, 'z',
      "Z_axis type", "amp|db|phase" },
    { "ifft", 'i',
      POPT_ARG_NONE,
      &ifft, 'i',
      "Enable ifft (z_axis=real)", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**) argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* input_file(zip) [output_file(csv)]");

    //argument check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case 't':
            if (strcmp(tf_typeStr, "LOC") == 0 || strcmp(tf_typeStr, "loc") == 0) {
                tf_type = LOC;
            } else if (strcmp(tf_typeStr, "SEP") == 0 || strcmp(tf_typeStr, "sep") == 0) {
                tf_type = SEP;
            } else {
                harkio_Log_printf(harkio_Log_Error, "%s: %s", poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
                        "invalid string value");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case 'z':
            if (strcmp(z_axisStr, "amp") == 0) {
                z_axis = DISP_AMP;
            } else if (strcmp(z_axisStr, "db") == 0) {
                z_axis = DISP_DB;
            } else if (strcmp(z_axisStr, "phase") == 0) {
                z_axis = DISP_PHASE;
            } else {
                harkio_Log_printf(harkio_Log_Error, "%s: %s", poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
                        "invalid string value");
                exit(EXIT_FAILURE);
            }
            hasParsedOpt = true;
            break;
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s", poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get argment
    const char* infile = poptGetArg(optCon);

    if (!infile) {
        harkio_Log_printf(harkio_Log_Error, "missing input_file");
        exit(EXIT_FAILURE);
    }

    harkio_TransferFunction *tf = harkio_TransferFunction_fromFile(infile, NULL);

    const char* outfile = poptGetArg(optCon);

    int ret = writeCoordinateFromTF(outfile, z_axis, tf, tf_type, ifft);
    // harkio_TransferFunction_delete(&tf);
    if (ret == EXIT_SUCCESS) {
        fprintf(stderr, "SUCCESS\n");
        if (outfile)
            fprintf(stderr, " output file : %s\n", outfile);
    }

    // free poptContext
    poptFreeContext(optCon);

    return ret;
}

/*---------------------------------------*/
// Z軸計算
/*---------------------------------------*/
float calcZValue(int z_unit, float re, float im) {
    switch (z_unit) {
    case DISP_REAL:
        return re;
        break;

    case DISP_AMP:
        return sqrtf(re * re + im * im);
        break;

    case DISP_DB:
        return log10(sqrtf(re * re + im * im)) * 10;
        break;

    case DISP_PHASE:
        return atan2f(im, re) * 180 / pi;
        break;

    default:
        harkio_Log_printf(harkio_Log_Error, "z_unit error [%d]", z_unit);
        exit(EXIT_FAILURE);
        break;
    }

    return -1;
}

/*------------------------------------------------------*/
//  座標変換
/*------------------------------------------------------*/
int writeCoordinateInternal(const char* outfile, int z_axis, harkio_Matrix** matdata, const harkio_Positions* poses,
        int num_mic, int num_x, float x_weight, const float* x_offset, float y_weight, const float* z_offset) {
    int num_src = poses->size;
    int micPosIds[num_mic];
    if (num_src > 0 && poses->pos[0]->channelsUse == 1) {
        num_mic = poses->pos[0]->numChannels;
        for (int i = 0; i < num_mic; i++) {
            micPosIds[i] = poses->pos[0]->channels[i];
        }
    } else {
        for (int i = 0; i < num_mic; i++) {
            micPosIds[i] = i;
        }
    }
    //printf("writeCoordinate start\n");

    FILE *fpw;
    float z_data;
    int src;
    int mic;
    int x;

    if (!outfile)
        fpw = stdout;
    else if ((fpw = fopen(outfile, "w")) == NULL) {
        harkio_Log_printf(harkio_Log_Error, "fopen error [%s]", outfile);
        exit(EXIT_FAILURE);
    }

    //write header
    fprintf(fpw, "id,mic,x,y,z\n");

    for (src = 0; src < num_src; src++) {
        for (mic = 0; mic < num_mic; mic++) {
            for (x = 0; x < num_x; x++) {
                // x_axis
                float x_data = x * x_weight + ((x_offset == NULL) ? 0.0 : x_offset[src]); // TIME[s] or FREQ
                // y_axis
                float y_data = src * y_weight;

                // z_axis
                COMPLEX_FLOAT z_complex_data = matdata[src]->data.complexdata[mic * num_x + x];
                if (z_offset) {
                    z_complex_data *= cexpf(-z_offset[src] * x * _Complex_I);
                }
                z_data = calcZValue(z_axis, crealf(z_complex_data), cimagf(z_complex_data));

                fprintf(fpw, "%d , %d , %f , %f , %f\n", poses->pos[src]->id, micPosIds[mic], x_data, y_data, z_data);
            }
        }                //mic for
    }                //src for

    fclose(fpw);

    return EXIT_SUCCESS;
}

void calcIFFTMatrixData(harkio_Matrix** freqdata, harkio_Matrix** timedata, int num_src, int num_mic, int num_fft) {
    int cnt;
    int src;
    int mic;
    int fft;

    COMPLEX_FLOAT* buff_A;
    float* float_a;

    buff_A = (COMPLEX_FLOAT*) malloc((num_fft / 2 + 1) * sizeof(COMPLEX_FLOAT));
    if (buff_A == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }
    float_a = (float*) malloc(num_fft * sizeof(float));
    if (float_a == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    // iFFTの実行
    for (src = 0; src < num_src; src++) { // f_max=g.Nsrc )
        for (mic = 0; mic < num_mic; mic++) {
            for (fft = 0; fft < num_fft / 2 + 1; fft++) {
                cnt = fft + mic * (num_fft / 2 + 1);
                buff_A[fft] = freqdata[src]->data.complexdata[cnt];
            }

            // fft実行 A[]の値をifftした結果がa[]にセットされる
            // Aのサイズ(複素対称)[0:Nfft/2] --> aのサイズ(実数のみ)[0:Nfft-1]
            // harktool_calcFloatIFFT( buff_A , float_a );
            harktool_FFTImpl_fRealFFTInv(buff_A, float_a, num_fft);

            for (fft = 0; fft < num_fft; fft++) {
                cnt = fft + mic * (num_fft);
                timedata[src]->data.complexdata[cnt] = (COMPLEX_FLOAT) float_a[fft];
            }
        }
    }
    free(float_a);
}

int getOffset(enum TFType tf_type, const harkio_TransferFunction* tf, float* offset) {
    char* srcPath = "transferFunction/source.xml";
    char* strSrcXml = NULL;
    if (as_zip_fread(tf->handle, srcPath, &strSrcXml) != EXIT_SUCCESS) {
        return EXIT_FAILURE;
    }
    // parse and modify source.xml
    xmlInitParser();
    xmlDocPtr doc = xmlParseDoc(BAD_CAST strSrcXml);
    free(strSrcXml);
    if (doc == NULL) {
        harkio_Log_printf(harkio_Log_Error, "source.xml parse error");
        return EXIT_FAILURE;
    }
    xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
    if (xpathCtx == NULL) {
        harkio_Log_printf(harkio_Log_Error, "xmlXPathNewContext error");
        xmlFreeDoc(doc);
        return EXIT_FAILURE;
    }
    for (int i = 0; i < tf->poses->size; i++) {
        char xpathExpr[100];
        sprintf(xpathExpr, "/hark_xml/positions/position[@id=%d]/calcTF[@type=%s]/@offset", tf->poses->pos[i]->id,
                (tf_type == LOC) ? "\"localization\"" : "\"separation\"");
        xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression(BAD_CAST xpathExpr, xpathCtx);
        if (xpathObj == NULL) {
            harkio_Log_printf(harkio_Log_Error, "xmlXPathEvalExpression error");
            xmlXPathFreeContext(xpathCtx);
            xmlFreeDoc(doc);
            return EXIT_FAILURE;
        }
        if (xpathObj->nodesetval != NULL && xpathObj->nodesetval->nodeNr > 0) {
            char* strOffset = (char*) xmlNodeGetContent(xpathObj->nodesetval->nodeTab[0]);
            offset[i] = (strOffset == NULL) ? 0.0 : atof((char*) strOffset);
            xmlFree(strOffset);
            if (tf_type == SEP) {
                offset[i] *= -1.0;
            }
        } else {
            offset[i] = 0.0;
        }
        xmlXPathFreeObject(xpathObj);
    }
    xmlXPathFreeContext(xpathCtx);
    xmlFreeDoc(doc);
    xmlCleanupParser();

    return EXIT_SUCCESS;
}

int writeCoordinateFromTF(const char *outfile, int z_axis, harkio_TransferFunction* tf, enum TFType tf_type, bool ifft) {
    float x_weight;
    float y_weight = 0.2;
    int x_num;

    if (!ifft) {
        x_weight = (float) tf->cfg->samplingRate / tf->cfg->nfft;
        x_num = tf->cfg->nfft / 2 + 1;
    } else {
        x_weight = 1.0 / tf->cfg->samplingRate;
        x_num = tf->cfg->nfft;
    }

    harkio_Matrix** matdata = NULL;

    if (tf_type == LOC) {
        matdata = tf->loctfs;
    } else if (tf_type == SEP) {
        matdata = tf->septfs;
    } else {
        harkio_Log_printf(harkio_Log_Error, "given tf_type");
        return EXIT_FAILURE;
    }
    if (matdata == NULL) {
        harkio_Log_printf(harkio_Log_Error, "no data");
        return EXIT_FAILURE;
    }

    float offset[tf->poses->size];
    if (z_axis == DISP_REAL || z_axis == DISP_PHASE || ifft) {
        // get offset
        if (getOffset(tf_type, tf, offset) == EXIT_FAILURE) {
            return EXIT_FAILURE;
        }
    }

    if (!ifft) {
        if (z_axis == DISP_REAL || z_axis == DISP_PHASE) {
            // convert offset(sample) -> delta_theta
            int nfft = (tf->cfg->nfft > 0) ? tf->cfg->nfft : 512;
            for (int src = 0; src < tf->poses->size; src++) {
                offset[src] *= 2.0 * pi / nfft;
            }
            return writeCoordinateInternal(outfile, z_axis, matdata, tf->poses, matdata[0]->rows, x_num, x_weight, NULL,
                    y_weight, offset);
        } else {
            return writeCoordinateInternal(outfile, z_axis, matdata, tf->poses, matdata[0]->rows, x_num, x_weight, NULL,
                    y_weight, NULL);
        }
    } else {
        // convert offset(sample) -> time
        for (int i = 0; i < tf->poses->size; i++) {
            offset[i] /= (tf->cfg->samplingRate > 0) ? tf->cfg->samplingRate : 16000.0;
        }
        // 結果格納用の行列
        harkio_Matrix* ifft_result[tf->poses->size];
        for (int src = 0; src < tf->poses->size; src++) {
            ifft_result[src] = harkio_Matrix_new("complex", tf->mics->size, tf->cfg->nfft);
            if (ifft_result[src] == NULL) {
                harkio_Log_printf(harkio_Log_Error, "memory error for ifft_result matrix");
                exit(EXIT_FAILURE);
            }
        }
        // data get
        calcIFFTMatrixData(matdata, ifft_result, tf->poses->size, tf->mics->size, tf->cfg->nfft);
        // file write
        writeCoordinateInternal(outfile, DISP_REAL, ifft_result, tf->poses, ifft_result[0]->rows, x_num, x_weight,
                offset, y_weight, NULL);
        // finalize
        for (int src = 0; src < tf->poses->size; src++) {
            harkio_Matrix_delete(&ifft_result[src]);
        }
        return EXIT_SUCCESS;
    }
}
